@@ -15,13 +15,33 @@
 (*  Based on Mldvi by Alexandre Miquel.                                *)
 (***********************************************************************)
 
-(* $Id: dviview.ml,v 1.3 2008/12/01 08:45:41 remy Exp $ *)
+(* $Id: dviview.ml,v 1.11 2011-08-02 07:21:10 remy Exp $ *)
 
 let pauses =
   Options.flag true "-nopauses"
   "  switch pauses off,\
   \n\t (the default is to wait for specified pauses).";;
 
+let events =
+  Options.add
+  "-pushkeys"
+    (Arg.String (String.iter Misc.push_char_event))
+  "<string>: push string chars as events\n\t";;
+
+
+let postsyncing =
+  Options.flag false "--postsyncing"
+  "  redraw with syncing if needed \n\t.";;
+
+let center_on_cursor =
+  Options.flag false "--center-on-cursor"
+  "  redraw so that cursor is displayed\n\t.";;
+  
+
+let scroll_fast =
+  Options.flag false "--scrollfast"
+  "  scroll to next page instead of page_down \n\t.";;
+
 let without_pauses f x =
   let p = !pauses in
   try pauses := false; let v = f x in pauses := p; v
@@ -591,7 +611,35 @@ let move_within_margins_x st movex =
     end in
   if st.orig_x <> new_orig_x then Some new_orig_x else None;;
 
-let redraw ?trans ?chst st =
+let make_visible st x y =
+  let hskip = 15 in
+  let vskip = 15 in
+  let movex =
+    let move_within_margins_x st dx = move_within_margins_x st dx in
+    let posx = x + st.orig_x in
+    if posx < 0 then
+      (* move far left to make x visible *)
+      (* move_within_margins_x st (attr.geom.Ageometry.width - 10) *)
+      move_within_margins_x st (attr.geom.Ageometry.width - hskip - posx)
+    else if posx > attr.geom.Ageometry.width then
+      (* move far right to make x visible *)
+      (* move_within_margins_x st (10 - attr.geom.Ageometry.width) *)
+      move_within_margins_x st (hskip - posx)
+    else None in
+  let movey = 
+    let posy = st.orig_y + y in
+    if posy < 0 then
+      move_within_margins_y st (attr.geom.Ageometry.height - vskip - posy)
+    else if posy > attr.geom.Ageometry.height then
+      move_within_margins_y st (vskip - posy)
+    else None  in
+  let r = false in
+  let r = match movex with | None -> r | Some dx -> st.orig_x <- dx; true in
+  let r = match movey with | None -> r | Some dy -> st.orig_y <- dy; true in
+  if r then set_bbox st;
+  r
+
+let rec redraw ?trans ?chst st =
   (* Draws until the current pause_number or page end. *)
   (* The pauses and waits that appear before are ignored. *)
   Busy.set Busy.Busy;
@@ -639,6 +687,13 @@ let redraw ?trans ?chst st =
   end;
   synchronize st;
   Busy.set (if st.cont = None then Busy.Free else Busy.Pause);
+  if not st.aborted &&  !postsyncing && not (Grdev.syncing()) then
+    begin
+      match st.dvi.Cdvi.pages.(st.page_number).Cdvi.page_status with
+      | Cdvi.Known { Cdvi.hasps = true } -> 
+          Grdev.with_syncing (redraw ?trans ?chst) st
+      | _ -> ()
+    end;
   Misc.debug_stop "Page has been drawn\n";;
 
 let goto_previous_pause n st =
@@ -753,13 +808,25 @@ let find_xref_here tag st =
   with Misc.Match ->
     Hashtbl.find (xrefs st) tag;;
 
+let scan_find_anchor_position st page html =
+  try
+    let  (x, y, _z, _w, _h) =
+      Driver.scan_find_anchor_position st.cdvi (st.base_dpi *. st.ratio)
+        page html in
+    Some (x, y)
+  with Not_found -> None 
+
 let page_start default st =
   match !start_html with
-  | None -> default
+  | None -> default, None
   | Some html ->
       Driver.scan_special_pages st.cdvi max_int;
-      try find_xref_here html st
-      with Not_found -> default;;
+      try
+        let page = find_xref_here html st in
+        let pos = scan_find_anchor_position st page html in
+        page, pos
+      with Not_found -> default, None;;
+
 
 (* foreground if drawing is needed after reloading *)
 let rec reload foreground st =
@@ -774,16 +841,21 @@ let rec reload foreground st =
     st.num_pages <- npages;
     st.toc <- None;
     st.page_stack <- clear_page_stack npages st.page_stack;
-    let npage = page_start (min st.page_number (st.num_pages - 1)) st in
+    let npage, pos = page_start (min st.page_number (st.num_pages - 1)) st in
     if npage <> st.page_number then
       begin
         st.pause_number <- 0;
         st.exchange_page <- st.page_number;
       end;
     set_page_number st npage;
+    begin if !center_on_cursor then
+      match pos with
+      | Some (x, y) -> ignore (make_visible st x y)
+      | None -> ()
+    end;
     st.frozen <- true;
     st.aborted <- true;
-    match st.duplex with
+    begin match st.duplex with
     | Client st' when not (compatible st st') ->
         Misc.warning
           (Printf.sprintf
@@ -801,6 +873,8 @@ let rec reload foreground st =
         update_dvi_size false st;
         Gs.init_do_ps ();
         if foreground then redraw ?trans:(Some Transitions.DirTop) st
+    end;
+    ()
   with x ->
     Misc.warning
       (Printf.sprintf "exception while reloading %s" (Printexc.to_string x));
@@ -824,8 +898,13 @@ let find_xref tag default st =
 
 let redisplay st =
   st.pause_number <- 0;
+  Grdev.toggle_syncing ();
   redraw st;;
 
+let resyncing st =
+  st.pause_number <- 0;
+  Grdev.with_syncing redraw st;;
+
 let show_thumbnails st r page =
   let size_x = Graphics.size_x () in
   let size_y = Graphics.size_y () in
@@ -855,7 +934,12 @@ let rec show_toc st =
       let roll = st.num mod n in
       st.next_num <- succ st.num;
       begin match rolls.(roll) with
-      | Page p -> redraw { st' with page_number = p }
+      | Page p -> redraw
+            { st' with
+              page_number = p;
+              orig_x = 0;
+              orig_y = top_of_page st;
+            }
       | Thumbnails (r, page) -> show_thumbnails st' r page
       end;
       synchronize st;
@@ -982,24 +1066,32 @@ let find_page_before_position (pos, filename as location) st =
   | [] -> raise Not_found
   | (p, _)::_ -> p
 
+let start_anchor = "Start-Document"
+
 let duplex_switch sync st =
   let find_sync_page master client =
-   let anchor = "Start-Document" in
-   match
-   (* master.dvi.Cdvi.pages.(master.page_number).Cdvi.line *)
-     Driver.scan_find_anchor_location master.cdvi 
-        (Hashtbl.find (xrefs master) anchor) anchor 
-   with
-    | None -> raise Not_found
-    | Some pos -> 
-   find_page_before_position pos client 
-  in
+    match
+      (* master.dvi.Cdvi.pages.(master.page_number).Cdvi.line *)
+      Driver.scan_find_anchor_location master.cdvi 
+        (Hashtbl.find (xrefs master) start_anchor) start_anchor 
+    with
+    | Some (pos, _) -> pos
+    | None -> raise Not_found in
 
   match st.duplex with
   | Alone -> ()
   | (Client st' | Master st') when not sync -> raise (Duplex (redraw, st'))
   | Client master ->
-      (try goto_page (find_sync_page master st) st with Not_found -> ())
+      (try
+        let page  = find_sync_page master st in
+        begin
+          (* should only be done if not all window shown *)
+          match scan_find_anchor_position st page start_anchor 
+          with
+          | Some (x, y) -> ignore (make_visible st x y); 
+          | None -> ()
+        end;
+      with Not_found -> ())
   | Master client ->
       try
         if changed client then reload false client;
@@ -1150,9 +1242,10 @@ let set_keymap, get_keymap =
 
 module B =
   struct
-    let nop c st =
+    let nop st = ()
+    let unbound c st =
       Misc.warning (Printf.sprintf "Unbound key %C" c)
-    let control_x_nop c st =
+    let control_x_unbound c st =
       set_keymap Default_keymap;
       Misc.warning (Printf.sprintf "Unbound Control-x key %C" c)
     let set_control_x_keymap st =
@@ -1184,6 +1277,7 @@ module B =
           (st.page_number - max 1 st.num) (st.last_pause_number) st
     let pop_page st =
       pop_page true 1 st
+
     let first_page st =
       goto_page 0 st
     let last_page st =
@@ -1270,6 +1364,14 @@ module B =
           end
       | None -> none ()
 
+    let scroll_down st =
+      if !scroll_fast then next_page st else page_down st
+    let scroll_up st =
+      if !scroll_fast then previous_page st else page_up st
+    let toggle_center_on_cursor = Options.toggle center_on_cursor
+
+    let scroll_switch st = scroll_fast := not !scroll_fast
+
     let redraw = redraw ?trans:(Some Transitions.DirNone) ?chst:None
 
     let toggle_active st =
@@ -1277,6 +1379,7 @@ module B =
 
     let reload = reload true
     let redisplay = redisplay
+    let resyncing = resyncing
 
     let toggle_full_screen st =
       let b = (st.full_screen = None) in
@@ -1315,10 +1418,10 @@ module B =
       Grdev.clean_ps_cache ()
     let help st =
       ignore (
-        Launch.fork_me
-          (Printf.sprintf "-g %dx%d"
-             attr.geom.Ageometry.width
-             attr.geom.Ageometry.height)
+      Launch.fork_me
+        (Printf.sprintf "-g %dx%d"
+           attr.geom.Ageometry.width
+           attr.geom.Ageometry.height)
         Config.splash_screen)
 
     let toggle_antialiasing st =
@@ -1326,7 +1429,7 @@ module B =
 
     let scratch_draw st =
       Scratch.draw ()
- 
+        
     let scratch_write st =
       Scratch.write ()
 
@@ -1338,7 +1441,7 @@ module B =
     let make_thumbnails st =
       Launch.without_launching
         (Busy.busy_exec
-          (fun () ->
+           (fun () ->
              make_thumbnails st;
              if not st.aborted then show_toc st))
         ()
@@ -1349,26 +1452,26 @@ module B =
     let ask_to_search =
       let prefill = ref "" in
       (fun message ->
-         let minibuff =
-           let nlines = 1 in
-           let xc, yc = 2, 2 in
-           let sx, sy = Graphics.text_size "X" in
-           let wt, ht = Graphics.size_x () - 2 * xc, sy * nlines in
-           let ncol = wt / sx in
-           let bw = 0 in
-           Gterm.make_term_gen
-             Graphics.black (Graphics.rgb 180 220 220) (* Grounds colors *)
-             bw Graphics.black (* Border *)
-             (Graphics.rgb 220 150 120) (* Title color *)
-             (Graphics.black) (* Cursor color *)
-             xc yc (* Where on the screen *)
-             ncol nlines (* Size in columns and lines *) in
-         Gterm.draw_term minibuff;
-         let re = Gterm.ask_prefill minibuff message !prefill in
-         (* Forces future redisplay. *)
-         Misc.push_key_event '' GraphicsY11.control;
-         prefill := re;
-         re)
+        let minibuff =
+          let nlines = 1 in
+          let xc, yc = 2, 2 in
+          let sx, sy = Graphics.text_size "X" in
+          let wt, ht = Graphics.size_x () - 2 * xc, sy * nlines in
+          let ncol = wt / sx in
+          let bw = 0 in
+          Gterm.make_term_gen
+            Graphics.black (Graphics.rgb 180 220 220) (* Grounds colors *)
+            bw Graphics.black (* Border *)
+            (Graphics.rgb 220 150 120) (* Title color *)
+            (Graphics.black) (* Cursor color *)
+            xc yc (* Where on the screen *)
+            ncol nlines (* Size in columns and lines *) in
+        Gterm.draw_term minibuff;
+        let re = Gterm.ask_prefill minibuff message !prefill in
+        (* Forces future redisplay. *)
+        Misc.push_key_event '' GraphicsY11.control;
+        prefill := re;
+        re)
 
     let search_forward st =
       let re_string = ask_to_search "Search Forward (re): " in
@@ -1403,7 +1506,7 @@ module B =
 
 let bind_key tbl (key, action) = tbl.(int_of_char key) <- action;;
 
-let default_keymap = Array.init 256 (fun i -> B.nop (char_of_int i));;
+let default_keymap = Array.init 256 (fun i -> B.unbound (char_of_int i));;
 let bind_default_key = bind_key default_keymap;;
 
 let bind_default_keys () =
@@ -1424,6 +1527,9 @@ let bind_default_keys () =
    'j', B.page_down;
    'k', B.page_up;
    'l', B.page_right;
+   '', B.toggle_center_on_cursor; 
+   (* toggle scroll fast *)
+   '\010', B.scroll_switch;
 
    (* m, i are reserved for scrolling
      'm', B.scroll_one_line_down;
@@ -1432,8 +1538,9 @@ let bind_default_keys () =
 
    (* return, tab, backspace, escape, and x
       to handle the page marks stack. *)
-   '\t' (* tab *), B.push_page;
-   '' (* Escape *), B.pop_page;
+   (* '\t' (* tab *), B.push_page; *)
+   '\t' (* tab *), B.show_toc;
+   '' (* Escape *), (*B.nop *) B.pop_page; 
 
    '\b' (* backspace *), B.pop_previous_page;
    '\r' (* return *), B.push_next_page;
@@ -1465,6 +1572,7 @@ let bind_default_keys () =
    'r', B.redraw;
    'R', B.reload;
    '', B.redisplay;
+   '/', B.resyncing;
 
    (* Control-f, c, to handle the advi window. *)
    '', B.toggle_full_screen;
@@ -1511,7 +1619,7 @@ let bind_default_keys () =
   ];;
 
 let control_x_keymap =
-  Array.init 256 (fun i -> B.control_x_nop (char_of_int i));;
+  Array.init 256 (fun i -> B.control_x_unbound (char_of_int i));;
 let bind_control_x_key = bind_key control_x_keymap;;
 
 (* Bindings for ^X-prefixed keys. *)
@@ -1560,7 +1668,7 @@ let main_loop mastername clients =
       Graphics.set_window_title  ("Advi: " ^ st.filename);
       if st.page_number > 0 && Gs.get_do_ps () then
         Driver.scan_special_pages st.cdvi st.page_number
-      else set_page_number st (page_start 0 st);
+      else set_page_number st (fst (page_start 0 st));
       if changed st then reload false st;
       idraw st;
       (* num is the current number entered by keyboard *)
@@ -1608,8 +1716,8 @@ let main_loop mastername clients =
         | Grdev.Position (x, y) ->
             st.last_command <- Position;
             position st x y
-        | Grdev.Click (pos, Grdev.Button4, _, _)  -> B.previous_page st
-        | Grdev.Click (pos, Grdev.Button5, _, _)  -> B.next_page st
+        | Grdev.Click (_, Grdev.Button4, _, _)  -> B.scroll_up st
+        | Grdev.Click (_, Grdev.Button5, _, _)  -> B.scroll_down st
         | Grdev.Click (pos, but, _, _) when Grdev.E.editing () ->
             begin match pos, but with
             | Grdev.Top_left, Grdev.Button1 -> B.previous_slice st
@@ -1629,6 +1737,7 @@ let main_loop mastername clients =
         | Grdev.Click (_, Grdev.Button3, _, _) ->
             if !click_turn_page then B.next_pause st
         | Grdev.Nil -> ()
+        | Grdev.Click (_, Grdev.NoButton, _, _)  -> ()
       done with
       | Exit -> Grdev.close_dev ()
       | Duplex (action, st') -> duplex action st' in
