@@ -11,7 +11,7 @@
 /*                                                                     */
 /***********************************************************************/
 
-/* $Id: grY11.c,v 1.6 2009/11/10 13:52:49 remy Exp $ */
+/* $Id: grY11.c,v 1.9 2010-08-18 14:47:12 remy Exp $ */
 
 #include <caml/memory.h>
 #include <caml/alloc.h>
@@ -23,6 +23,12 @@
 #include "libgraph.h"
 #include "image.h"
 
+#include <sys/types.h>
+#include <sys/time.h>
+#include <string.h>
+#include <unistd.h>
+
+
 #include "grwm.h"
 /* the HAVE_XINERAMA variable should be defined by configure */
 /* #define HAVE_XINERAMA */
@@ -391,7 +397,7 @@ value caml_gr_get_modifiers(void)
                     &root_x, &root_y, &win_x, &win_y,
                     &modifiers)) {
     button = 0;
-/*     fprintf(stderr,"C modifiers = %u\n", modifiers); */
+    /*     fprintf(stderr,"C modifiers = %u\n", modifiers); */
     if (modifiers & Button1Mask) button = button | 0x1;
     if (modifiers & Button2Mask) button = button | 0x2;
     if (modifiers & Button3Mask) button = button | 0x4;
@@ -612,3 +618,27 @@ value caml_gr_rebind_keysyms(value unit)
   return Val_unit;
 }
 
+
+value caml_gr_get_button(value m)
+{
+  int modifiers = Int_val(m);
+  int button = 0;
+  if (modifiers & Button1Mask) button = 1;
+  else if (modifiers & Button2Mask) button = 2;
+  else if (modifiers & Button3Mask) button = 3;
+  else if (modifiers & Button4Mask) button = 4;
+  else if (modifiers & Button5Mask) button = 5;
+  return Val_int(button);
+}
+
+struct event_data {
+  short kind;
+  short mouse_x, mouse_y;
+  unsigned char button;
+  unsigned char key;
+};
+
+extern struct event_data caml_gr_queue[SIZE_QUEUE];
+extern unsigned int caml_gr_head;       /* position of next read */
+extern unsigned int caml_gr_tail;       /* position of next write */
+
