@@ -15,7 +15,7 @@
 (*  Based on Mldvi by Alexandre Miquel.                                *)
 (***********************************************************************)
 
-(* $Id: grdev.ml,v 1.5 2009/04/03 17:21:29 remy Exp $ *)
+(* $Id: grdev.ml,v 1.15 2010-08-29 21:23:50 remy Exp $ *)
 
 open Misc;;
 
@@ -82,16 +82,61 @@ let flush_ps () =
   psused := true;
   Gs.flush ();;
 
-let flush_dvi () = GraphicsY11.flush ();;
-let flush_last () = if !last_is_dvi then flush_dvi () else flush_ps ();;
+let flush_dvi () =
+  GraphicsY11.flush ();;
+
+type syncing = Always | Page | Now | Later | Never
+let syncing_status = ref Never (* Immediately *)
+let reset_syncing () =
+  match !syncing_status with
+  | Page -> syncing_status := Never
+  | Now -> syncing_status := Later
+  | _ -> ()
+
+let toggle_syncing () =
+  match !syncing_status with 
+  | Always -> syncing_status := Never
+  | Never -> syncing_status := Always 
+  | _ -> ()
+
+let try_finalize f x finally y =
+  let res = try f x with exn -> finally y; raise exn in 
+  finally y; 
+  res
+
+let with_syncing f x =
+  let old = !syncing_status in
+  let cur = 
+  match old with
+  | Always | Page | Now as x -> x
+  | Later -> Now
+  | Never -> Page in
+  syncing_status := cur;
+  let res = try f x with exn -> syncing_status := old; raise exn in
+  syncing_status := old;
+  res 
+
+let syncing() =
+  match !syncing_status with
+  | Always | Page | Now -> true
+  | Later | Never -> false
+
+let flush_last () =
+  if syncing() then 
+    if  !last_is_dvi then flush_dvi () else flush_ps ();;
 
 let sync b =
-  if !last_is_dvi <> b then begin flush_last (); last_is_dvi := b end;;
+  if syncing() then
+    if !last_is_dvi <> b then begin flush_last (); last_is_dvi := b end;;
+
+let dosync_ps() =
+  if !last_is_dvi <> true then begin flush_ps (); last_is_dvi := true end;;
 
 let title = ref "Advi";;
 let set_title s = title := s;;
 
 let synchronize () =
+  if syncing() then () else dosync_ps ();
   Gs.flush ();
   Transimpl.synchronize_transition ();
   GraphicsY11.synchronize ();
@@ -1333,7 +1378,8 @@ type status = GraphicsY11.status = {
 type area =
    | Bottom_right | Bottom_left | Top_right | Top_left | Middle;;
 type button =
-   | Button1 | Button2 | Button3 | Button4 | Button5;;
+   | Button1 | Button2 | Button3 | Button4 | Button5 | NoButton 
+
 type event =
    | Resized of int * int
    | Refreshed
@@ -1444,7 +1490,6 @@ let resized () =
     end
   else None;;
 
-
 let rec wait_signal_event events =
   match resized (), !usr1_status,  event_waiting (), !usr2_status with
   | Some (x, y), _, _, _ -> Final (Resized (x, y))
@@ -1630,44 +1675,57 @@ let mouse_area near x y =
     else Middle
   else Middle;;
 
-let pressed m b = m land b <> 0;;
-let released m b = m land b = 0;;
+let modifier m b = m land b <> 0;;
 
 module G = GraphicsY11;;
 
-let button m =
-  if pressed m G.button1 then Button1 else
-  if pressed m G.button3 then Button3 else
-  if pressed m G.button4 then Button4 else
-  if pressed m G.button5 then Button5 else
-  Button2;;
+let pressed m b = m land b <> 0
+let button123 = G.button1 lor G.button2 lor G.button3
+
+let get_button b =
+  if pressed b G.button1 then Button1 else
+  if pressed b G.button2 then Button2 else
+  if pressed b G.button3 then Button3 else
+  if pressed b G.button4 then Button4 else
+  if pressed b G.button5 then Button5 else
+  NoButton
+;;
+
+let shift_or_control = G.shift lor G.control
+
+let button_pressed m b = get_button m = b
 
 let wait_button_up m x y =
-  (*  Printf.printf  "Button %d\n%!" (Obj.magic (Obj.repr (button m))); *)
   let wait_position () =
     match wait_signal_event button_up with
     | Raw e ->
-        if !editing || pressed m G.shift then begin
+        let m = e.modifiers in
+        let () = Misc.debug_endline (Printf.sprintf "wait_position %x" m) in
+        if modifier m G.button1 && not (modifier m shift_or_control) then begin
           match mouse_area close x y with
           | Middle -> Final (Position (x, !size_y - y))
-          | c -> Final (Click (c, button m, x, !size_y - y))
+          | c -> Final (Click (c, get_button e.modifiers, x, !size_y - y))
         end
-        else Final (Click (mouse_area near x y, button m, x, !size_y - y))
+        else
+          let c = mouse_area near x y  in
+          Final (Click (c, get_button e.modifiers, x, !size_y - y))
     | x -> x in
-  if !editing && pressed m G.button1 then wait_position () else
-  if !editing || pressed m G.control then begin
+  if modifier m G.button1 && not (modifier m shift_or_control) then
+    wait_position () else
+  if modifier m G.control || modifier m G.shift then begin
     try
+      let () = Misc.debug_endline (Printf.sprintf "Control?") in
       let p = E.find x y in
       let rect = p.E.rect in
       let info = p.E.info in
       let action = info.E.action in
-      if pressed m G.button2 && (action.rx || action.ry) then
+      if button_pressed m Button2 && (action.rx || action.ry) then
         let event dx dy = Edit (p, E.Move (dx, dy)) in
         let action = { action with rw = false; rh = false; rd = false} in
         wait_move_button_up rect action event x y
-      else if (pressed m G.button3 || pressed m G.button1)
+      else if (button_pressed m Button3 || button_pressed m Button1)
           && (action.rh || action.rd || action.rw) then
-        let b = pressed m G.shift in
+        let b = modifier m G.shift in
         let action_h = action.rh && (not action.rd || not b) in
         let action_d = action.rd && (not action.rh || b) in
         let event dx dy = Edit (p, E.Resize (action_h, dx, dy)) in
@@ -1678,14 +1736,15 @@ let wait_button_up m x y =
       else Final Nil
     with
     | Not_found ->
-        if pressed m G.control then
+        let () = Misc.debug_endline (Printf.sprintf "Control not found?") in
+        if modifier m G.control then
           let event dx dy = Move (dx, 0-dy) in
           let move =
             { rx = true; ry = true; rw = false; rh = false; rd = false } in
           wait_move_button_up !bbox move event x y
         else wait_position ()
   end else
-    if pressed m G.shift && released m G.button1
+    if modifier m G.shift && not (button_pressed m Button1)
     then wait_select_button_up m x y
     else wait_position ();;
 
@@ -1693,7 +1752,7 @@ let wait_button_up m x y =
 let find_key ev =
   let c = ev.GraphicsY11.key in
   let k =
-    if ev.GraphicsY11.modifiers != GraphicsY11.control then c else
+    if ev.modifiers != G.control then c else
     let cc = 1 + int_of_char c - int_of_char 'A' in
     if cc >= 1 then char_of_int cc else c in
   Key k;;
@@ -1709,19 +1768,22 @@ let wait_event () =
     match wait_signal_event all_events with
     | Final e -> send e
     | Raw ev ->
-        if ev.GraphicsY11.keypressed then send (find_key ev) else
+        if ev.G.keypressed then send (find_key ev) else
         try match H.find_with_tag active_tag ev.mouse_x ev.mouse_y with
         | {A.action = {H.tag = H.Href h; H.draw = d}} as act ->
             if ev.button then
               let _ev' = GraphicsY11.wait_next_event button_up in
-              send (Href h) else
-              if H.up_to_date act emph then event emph b else begin
+              if  pressed ev.modifiers button123
+              then send (Href h)
+              else send (Click (Middle, get_button _ev'.modifiers,
+                                _ev'.mouse_x, _ev'.mouse_y))
+            else if H.up_to_date act emph then event emph b else begin
                 H.deemphasize true emph;
                 event (H.emphasize_and_flash href_emphasize_color act) b end
         | {A.action = {H.tag = H.Item s; H.draw = d}} as act ->
             if ev.button then
               let _ev' = GraphicsY11.wait_next_event button_up in
-              let area = 
+              let area =
                 match s with 
                 | "Bottom_left" -> Bottom_left
                 | "Bottom_right" -> Bottom_right
@@ -1729,8 +1791,9 @@ let wait_event () =
                 | "Top_right" -> Top_right 
                 | _ -> Middle
               in
+              (* This would not work with rolling button *)
               send (Click (area, 
-                           button (GraphicsY11.get_modifiers()), 
+                           get_button _ev'.modifiers, 
                            _ev'.mouse_x, 
                            _ev'.mouse_y))
             else
@@ -1764,29 +1827,29 @@ let wait_event () =
                   H.deemphasize true emph;
                   event (H.emphasize_and_flash href_emphasize_color act) b end
         | _ -> rescan ()
-                   with Not_found ->
-                     if ev.button then
-                       let m = GraphicsY11.get_modifiers () in
-                       match wait_button_up m ev.mouse_x ev.mouse_y with
-                       | Final (Region (x, y, dx, dy) as e) -> send e
-                       | Final (Selection s as e) -> send e
-                       | Final (Position (x, y) as e) -> send e
-                       | Final (Move (dx, dy) as e) -> send e
-                       | Final (Click (_, _, _, _) as e) -> send e
-                       | Final (Edit (_, _) as e) -> send e
-                       | Final Nil -> send Nil
-                       | Final
-                           (Resized (_, _) | Refreshed |
-                           Key _ | Stdin _ | Href _ | Advi (_, _) as e) ->
-                             push_back_event ev;
-                             send e
-                       | Raw _ -> rescan ()
-                     else
-                       let m = GraphicsY11.get_modifiers () in
-                       if m land GraphicsY11.shift <> 0
-                       then Busy.temp_set Busy.Selection
-                       else Busy.restore_cursor ();
-                       rescan () in
+        with Not_found ->
+          if ev.button then
+            let m = ev.modifiers in
+            match wait_button_up m ev.mouse_x ev.mouse_y with
+            | Final (Region (x, y, dx, dy) as e) -> send e
+            | Final (Selection s as e) -> send e
+            | Final (Position (x, y) as e) -> send e
+            | Final (Move (dx, dy) as e) -> send e
+            | Final (Click (_, _, _, _) as e) -> send e
+            | Final (Edit (_, _) as e) -> send e
+            | Final Nil -> send Nil
+            | Final
+                (Resized (_, _) | Refreshed |
+                Key _ | Stdin _ | Href _ | Advi (_, _) as e) ->
+                  push_back_event ev;
+                  send e
+            | Raw _ -> rescan ()
+          else
+            let m = GraphicsY11.get_modifiers () in
+            if not (modifier m G.shift)
+            then Busy.temp_set Busy.Selection
+            else Busy.restore_cursor ();
+            rescan () in
   event H.Nil false;;
 
 (* To be changed *)
@@ -1797,15 +1860,17 @@ let resized () =
 
 let continue () =
   if
-    resized () || 
-    GraphicsY11.key_pressed () (*  || !usr1_status *) ||
+    resized () || (*  !usr1_status || *)
+    GraphicsY11.key_pressed () ||
+    (* button pressed cannot be implemented as an extension to Graphics *)
+    GraphicsY11.button_pressed () ||
     !usr2_status (* input from stdin *)
   then begin Gs.flush (); raise Stop end;;
 
 (* Calling GS *)
 
 let current_pos () =
-  if not !last_is_dvi then flush_ps ();
+  if syncing() && not !last_is_dvi then flush_ps ();
   let x, y = Gs.current_point() in
   x, !size_y - y;;
 
